//
// turing - simulation of Turing machines
// Copyright (C) 2009  Joachim Reichel <joachim.reichel@gmx.de>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//

/// \file
/// \brief The #TM::Alphabet class.

#ifndef TURING_ALPHABET_H
#define TURING_ALPHABET_H

#include "types.h"

#include <string>
#include <ostream>

namespace TM {



/// Set of available characters
///
/// An alphabet indicates the set of characters available for a Turing machine.
/// The alphabet always has to include the special character #Blank.
///
/// The characters of an alphabet are ordered. The index of an character in
/// this order is used to encode the character in a Goedel number.
class Alphabet {

public:
    /// Constructor
    ///
    /// The default alphabet consists of the characters '0', '1', and
    /// 'B' (#Blank).
    Alphabet( const Character* alphabet = "01B");

    /// Returns the index of a character
    ///
    /// \param character   The character to be looked up.
    /// \return            The index of \p character, or -1 if the alphabet
    ///                    does not contain this character.
    size_t get_character_index( Character character) const;

    /// Returns the character with a given index
    ///
    /// \pre               \p index < #get_nr_of_characters()
    ///
    /// \param index       The index of the character to return.
    /// \return            The character that has index \p index.
    inline Character get_character( size_t index) const
    {
        TM_EXPENSIVE_CHECK( index < m_alphabet.size());
        return m_alphabet[index];
    }

    /// Retunrs the number of characters of the alphabet
    size_t get_nr_of_characters() const { return m_alphabet.size(); }

    /// Equality operator
    bool operator==( const Alphabet& rhs) const;

    /// Inequality operator
    bool operator!=( const Alphabet& rhs) const;

    /// Output operator
    friend std::ostream& operator<<( std::ostream& out, const Alphabet& tape);

private:
    std::string m_alphabet;
    size_t m_character_index[256];
};

/// Output operator
std::ostream& operator<<( std::ostream& out, const Alphabet& tape);



} // namespace TM

#endif // TURING_ALPHABET_H
