//
// turing - simulation of Turing machines
// Copyright (C) 2009  Joachim Reichel <joachim.reichel@gmx.de>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//

#include "goedel.h"

#include <cstdlib>
#include <sstream>

#ifndef NDEBUG
#  include <iostream>
#endif

namespace TM {



namespace internal {

// Encode a transition for encoding F0 or F1
std::string goedel_transition_01(
    const Transition& transition, Encoding encoding, const Alphabet& alphabet)
{
    TM_EXPENSIVE_CHECK(
        ( encoding == ENCODING_F0) || ( encoding == ENCODING_F1));
    // The F1 encoding does not support transitions form q != 0 to q' == 0
    TM_EXPENSIVE_CHECK(
        ( encoding != ENCODING_F1) ||
        ( transition.get_state() == 0) || ( transition.get_new_state() != 0));
    // The F1 encoding does not support transitions form q == 0 to q' == 1
    TM_EXPENSIVE_CHECK(
        ( encoding != ENCODING_F1) ||
        ( transition.get_state() != 0) || ( transition.get_new_state() != 1));

    State q = transition.get_state();
    Character c = transition.get_character();
    State Q = transition.get_new_state();
    Character C = transition.get_new_character();
    Direction d = transition.get_direction();

    std::string s;
    s += std::string( q + 1, '0');
    s += "1";
    s += std::string( alphabet.get_character_index( c) + 1, '0');
    s += "1";
    s += std::string( (encoding == ENCODING_F1) && ( Q == q + 1) ? 1 : Q + 1, '0');
    s += "1";
    s += std::string( alphabet.get_character_index( C) + 1, '0');
    s += "1";
    s += std::string( ( d == Left) ? 1 : (( d == Neutral) ? 2 : 3), '0');
    return s;
}

// Encode a transition for encoding F2
std::string goedel_transition_2(
    const Transition& transition, Encoding encoding, const Alphabet& alphabet)
{
    TM_EXPENSIVE_CHECK( encoding == ENCODING_F2);
    // The F1 encoding does not support transitions form q != 0 to q' == 0
    TM_EXPENSIVE_CHECK(
        ( transition.get_state() == 0) || ( transition.get_new_state() != 0));

    State q = transition.get_state();
    State Q = transition.get_new_state();
    Character C = transition.get_new_character();
    Direction d = transition.get_direction();

    if( ( Q == q+1) && ( d == Right))
        return std::string( alphabet.get_character_index( C) + 3, '0');

    std::string s;
    s += std::string( ( Q == q) ? 1 : Q + 1, '0');
    s += "1";
    s += std::string( alphabet.get_character_index( C) + 1, '0');
    s += "1";
    s += std::string( ( d == Right) ? 1 : (( d == Neutral) ? 2 : 3), '0');
    return s;
}

// Encode a Turing machine for encoding F0 or F1
size_t goedel_01(
    std::ostream& stream, const Turing_machine& turing_machine,
    Encoding encoding, bool incomplete)
{
    TM_ALWAYS_CHECK(
        ( encoding == ENCODING_F0) || ( encoding == ENCODING_F1));

    size_t size = 0;
    const Alphabet& alphabet = turing_machine.get_alphabet();

    for( size_t i = 0; i < turing_machine.get_nr_of_transitions(); ++i) {
        std::string s = internal::goedel_transition_01(
            turing_machine.get_transition( i), encoding, alphabet);
        s += "11";
        stream << s;
        size += s.size();
    }

    return size;
}

// Encode a Turing machine for encoding F2
size_t goedel_2(
    std::ostream& stream, const Turing_machine& turing_machine,
    Encoding encoding, bool incomplete)
{
    TM_ALWAYS_CHECK( encoding == ENCODING_F2);

    size_t size = 0;
    Turing_machine tm = turing_machine;
    tm.optimize();
    const Alphabet& alphabet = tm.get_alphabet();

    State q_max
        = tm.get_transition( tm.get_nr_of_transitions()-1 ).get_state();
    size_t c_max = alphabet.get_nr_of_characters()-1;
    size_t n_missing_transitions = 0;

    for( State q = 0; q <= q_max; ++q) {
        for( size_t c_i = 0; c_i <= c_max; ++c_i) {

            Character c = alphabet.get_character( c_i);
            size_t i = tm.find_transition( q, c);

            if( i != static_cast<size_t>( -1)) {
                // encode previous non-existing transitions from state q
                while( n_missing_transitions > 0) {
                    stream << "0011";
                    size += 4;
                    n_missing_transitions--;
                }
                // encode this transition
                std::string s = goedel_transition_2(
                    tm.get_transition( i), encoding, alphabet);
                s += "11";
                stream << s;
                size += s.size();
            } else
                // record non-existing transition
                n_missing_transitions += 1;
        }

        // encode previous non-existing transitions from state q
        if( n_missing_transitions > 0) {
            stream << "011";
            size += 3;
            n_missing_transitions = 0;
        }
    }

    return size;
}

// Returns length of longest prefix of zeroes.
//
// Assumes \p contains either '0' or '1'.
// Advances \p index by this length plus 1.
size_t decode( const std::string s, size_t& index)
{
    size_t pos = s.find( '1', index);
    if( pos == std::string::npos)
        pos = s.size();
    size_t len = pos - index;
    index = pos + 1;
    return len;
}

// Decode a transition for encoding F0 or F1
Transition* ungoedel_transition_01(
    const std::string& s, Encoding encoding, const Alphabet& alphabet)
{
    TM_EXPENSIVE_CHECK(
        ( encoding == ENCODING_F0) || ( encoding == ENCODING_F1));

    size_t index = 0;
    size_t n = decode( s, index);
    if( n == 0)
        return NULL;
    State q = n - 1;
    n = decode( s, index);
    if( n == 0)
        return NULL;
    Character c = alphabet.get_character( n - 1);
    n = decode( s, index);
    if( n == 0)
        return NULL;
    State Q = ( encoding == ENCODING_F0) || ( n > 1) ? n - 1 : (( q == 0) ? 0 : q + 1);
    n = decode( s, index);
    if( n == 0)
        return NULL;
    Character C = alphabet.get_character( n - 1);
    n = decode( s, index);
    if( ( n == 0) || ( n > 3))
        return NULL;
    Direction d = ( n == 1 ? Left : (( n == 2) ? Neutral : Right));
    if( index != s.size() + 1)
        return NULL;
    return new Transition( q, c, Q, C, d);
}

// Decode a transition for encoding F2
Transition* ungoedel_transition_2(
    const std::string& s, State q, Character c, Encoding encoding,
    const Alphabet& alphabet)
{
    TM_EXPENSIVE_CHECK( encoding == ENCODING_F2);

    if( s.find( "1") == std::string::npos) {
        Character C = alphabet.get_character( s.size() - 3);
        return new Transition( q, c, q+1, C, Right);
    }

    size_t index = 0;
    size_t n = decode( s, index);
    State Q = ( n > 1) ? n - 1 : q;
    n = decode( s, index);
    if( n == 0)
        return NULL;
    Character C = alphabet.get_character( n - 1);
    n = decode( s, index);
    if( ( n == 0) || ( n > 3))
        return NULL;
    Direction d = ( n == 1 ? Right : (( n == 2) ? Neutral : Left));
    if( index != s.size() + 1)
        return NULL;
    return new Transition( q, c, Q, C, d);
}

// Appends data from \p stream to \p s, until \p s contains "11" or the end of
// the stream is reached.
//
// Returns the position of "11" (or std::string::npos if the end of the stream
// is reached).
size_t find( std::istream& stream, std::string& s)
{
    size_t pos = s.find( "11");
    while( (pos == std::string::npos) && stream.good()) {
        char buffer[1048576+1];
        stream.read( buffer, 1048576);
        buffer[stream.gcount()] = '\0';
        s += buffer;
        pos = s.find( "11");
    }
    return pos;
}

// Decode a Goedel number for encoding F0 or F1
Turing_machine* ungoedel_01( std::istream& stream, Encoding encoding,
    const Alphabet& alphabet)
{
    TM_ALWAYS_CHECK(
        ( encoding == ENCODING_F0) || ( encoding == ENCODING_F1));

    std::string buffer;
    internal::find( stream, buffer);

    if( buffer.substr( 0, 3) != "111") {
#ifndef NDEBUG
        std::cout << "error (ungoedel failed, does not start with 111) "
                  << std::endl;
#endif
        return NULL;
    }
    buffer = buffer.substr( 3);

    Turing_machine* tm = new Turing_machine( 0, 1, alphabet);

    size_t end = internal::find( stream, buffer);
    while( end != std::string::npos) {

        std::string s = buffer.substr( 0, end);
        Transition* transition
            = internal::ungoedel_transition_01( s, encoding, alphabet);
        if( !transition) {
#ifndef NDEBUG
            std::cout << "error (ungoedel failed for this transition): "
                      << s << std::endl;
#endif
            delete tm;
            return NULL;
        }
        tm->add_transition( *transition);
        delete transition;
        buffer = buffer.substr( end + 2);
        end = internal::find( stream, buffer);
    }

    if( buffer != "1") {
#ifndef NDEBUG
        std::cout << "error (ungoedel failed, does not end with 111): 11"
                  << buffer << std::endl;
#endif
        delete tm;
        return NULL;
    }

    return tm;
}

// Decode a Goedel number for encoding F2
Turing_machine* ungoedel_2( std::istream& stream, Encoding encoding,
    const Alphabet& alphabet)
{
    TM_ALWAYS_CHECK( encoding == ENCODING_F2);

    std::string buffer;
    internal::find( stream, buffer);

    if( buffer.substr( 0, 3) != "111") {
#ifndef NDEBUG
        std::cout << "error (ungoedel failed, does not start with 111) "
                  << std::endl;
#endif
        return NULL;
    }
    buffer = buffer.substr( 3);

    Turing_machine* tm = new Turing_machine( 0, 1, alphabet);

    State q = 0;
    Character c = alphabet.get_character( 0);

    size_t end = internal::find( stream, buffer);
    while( end != std::string::npos) {

        std::string s = buffer.substr( 0, end);

        // no more transition for (q, .)
        if( s == "0") {
            q++;
            c = alphabet.get_character( 0);

        // no transition for (q, c)
        } else if( s == "00") {
            size_t c_i = ( alphabet.get_character_index( c) + 1)
                % alphabet.get_nr_of_characters();
            if( c_i == 0)
              q++;
            c = alphabet.get_character( c_i);

        // transition for (q, c)
        } else {
            Transition* transition
                = internal::ungoedel_transition_2(
                    s, q, c, encoding, alphabet);
            if( !transition) {
#ifndef NDEBUG
                std::cout << "error (ungoedel failed for this transition): "
                          << s << std::endl;
#endif
                delete tm;
                return NULL;
            }
            tm->add_transition( *transition);
            delete transition;

            size_t c_i = ( alphabet.get_character_index( c) + 1)
                % alphabet.get_nr_of_characters();
            if( c_i == 0)
              q++;
            c = alphabet.get_character( c_i);
        }

        buffer = buffer.substr( end + 2);
        end = internal::find( stream, buffer);
    }

    if( buffer != "1") {
#ifndef NDEBUG
        std::cout << "error (ungoedel failed, does not end with 111): 11"
                  << buffer << std::endl;
#endif
        delete tm;
        return NULL;
    }

    return tm;
}

} // namespace

std::string goedel(
    const Turing_machine& turing_machine, Encoding encoding, bool incomplete)
{
    std::ostringstream str;
    goedel( str, turing_machine, encoding, incomplete);
    return str.str();
}

size_t goedel(
    std::ostream& stream, const Turing_machine& turing_machine,
    Encoding encoding, bool incomplete)
{
    TM_ALWAYS_CHECK(
        incomplete || ( turing_machine.get_initial_state() == 0));
    TM_ALWAYS_CHECK(
        incomplete || ( turing_machine.get_halting_state() == 1));
    TM_ALWAYS_CHECK(
        ( encoding == ENCODING_F0)
     || ( encoding == ENCODING_F1)
     || ( encoding == ENCODING_F2));

    size_t size = 0;
    stream << "111";
    size += 3;

    if( encoding == ENCODING_F2)
        size += internal::goedel_2(
            stream, turing_machine, encoding, incomplete);
    else
        size += internal::goedel_01(
            stream, turing_machine, encoding, incomplete);

    stream << "1";
    size += 1;
    return size;
}

Turing_machine* ungoedel( const std::string& s, Encoding encoding,
    const Alphabet& alphabet)
{
    std::istringstream str( s);
    return ungoedel( str, encoding, alphabet);
}

Turing_machine* ungoedel( std::istream& stream, Encoding encoding,
    const Alphabet& alphabet)
{
    TM_ALWAYS_CHECK(
        ( encoding == ENCODING_F0)
     || ( encoding == ENCODING_F1)
     || ( encoding == ENCODING_F2));

    if( encoding == ENCODING_F2)
        return internal::ungoedel_2( stream, encoding, alphabet);
    else
        return internal::ungoedel_01( stream, encoding, alphabet);
}



} // namespace TM
