//
// turing - simulation of Turing machines
// Copyright (C) 2009  Joachim Reichel <joachim.reichel@gmx.de>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//

#include "tape.h"

namespace TM {



Tape::Tape( const Character* initial_content)
{
    TM_ALWAYS_CHECK( initial_content);

    size_t len = 0;
    while( initial_content[len]) {
        TM_ALWAYS_CHECK( initial_content[len] != Blank);
        ++len;
    }

    // avoid initial size 0 to avoid reallocation problems later
    m_right_tape.resize( (len > 0) ? len : 16, Blank);
    m_left_tape.resize( 16, Blank);

    ptrdiff_t index = 0;
    while( initial_content[index]) {
        m_right_tape[index] = initial_content[index];
        ++index;
    }

    m_left_boundary = 0;
    m_right_boundary = std::max(
        static_cast<ptrdiff_t>( 0), static_cast<ptrdiff_t>( len)-1);
    m_boundaries_uptodate = true;
}

bool Tape::is_valid( const Alphabet& alphabet) const
{
    for( ptrdiff_t index = get_left_boundary();
        index < get_right_boundary(); ++index)
        if( alphabet.get_character_index( get_character( index))
            == static_cast<size_t>( -1))
            return false;
    return true;
}

ptrdiff_t Tape::get_left_boundary() const
{
    if( !m_boundaries_uptodate)
        update_boundaries();
    return m_left_boundary;
}

ptrdiff_t Tape::get_right_boundary() const
{
    if( !m_boundaries_uptodate)
        update_boundaries();
    return m_right_boundary;
}

void Tape::update_boundaries() const
{
    while( (get_character( m_left_boundary ) == Blank)
        && (m_left_boundary < m_right_boundary))
        ++m_left_boundary;
    while( (get_character( m_right_boundary) == Blank)
        && (m_right_boundary > m_left_boundary))
        --m_right_boundary;

    TM_ALWAYS_CHECK( ( get_character( m_left_boundary)  != Blank)
         || ( m_left_boundary == m_right_boundary));
    TM_ALWAYS_CHECK( ( get_character( m_right_boundary) != Blank)
         || ( m_left_boundary == m_right_boundary));
    TM_ALWAYS_CHECK( get_character( m_left_boundary-1)  == Blank);
    TM_ALWAYS_CHECK( get_character( m_right_boundary+1) == Blank);

    m_boundaries_uptodate = true;
}

bool Tape::operator==( const Tape& rhs) const
{
    update_boundaries();
    ptrdiff_t rhs_left_boundary = rhs.get_left_boundary();
    ptrdiff_t rhs_right_boundary = rhs.get_right_boundary();
    if( m_left_boundary != rhs_left_boundary)
        return false;
    if( m_right_boundary != rhs_right_boundary)
        return false;
    for( ptrdiff_t i = m_left_boundary; i <= m_right_boundary; ++i)
        if( get_character( i) != rhs.get_character( i))
            return false;
    return true;
}

bool Tape::operator!=( const Tape& rhs) const
{
    return !( *this == rhs);
}

std::ostream& operator<<(std::ostream& stream, const Tape& tape)
{
    for( ptrdiff_t index = tape.get_left_boundary();
        index < tape.get_right_boundary(); ++index)
        stream << tape.get_character( index);
    // do not output B for the empty string
    Character c = tape.get_character( tape.get_right_boundary());
    if( c != Blank)
        stream << c;
    return stream;
}



} // namespace TM
