//
// turing - simulation of Turing machines
// Copyright (C) 2009  Joachim Reichel <joachim.reichel@gmx.de>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//

#include "turing.h"

#include <fstream>
#include <string>
#include <iostream>
#include <sstream>
#include <algorithm>

void print_stats(
    const std::string& prefix,
    const TM::Turing_machine& tm,
    TM::Encoding encoding)
{
   std::cout << prefix
       << " initial state: " << tm.get_initial_state() << std::endl;
   std::cout << prefix
       << " halting state: " << tm.get_halting_state() << std::endl;
   std::cout << prefix
       << " alphabet:      " << tm.get_alphabet() << std::endl;
   std::cout << prefix
       << " #states:       " << tm.get_nr_of_states() << std::endl;
   std::cout << prefix
       << " #transitions:  " << tm.get_nr_of_transitions() << std::endl;
   std::cout << prefix
       << " encoding:      " << encoding << std::endl;
   std::cout << std::endl;
}

TM::Turing_machine construct_tm_from_file(
    const std::string& filename,
    TM::State& initial_state_w,
    TM::State& offset_state_w,
    TM::Encoding& encoding)
{
    std::ifstream f( filename.c_str());
    TM::State initial_state_t;
    int encoding_int;
    std::string alphabet_string;
    f >> initial_state_w >> offset_state_w
      >> initial_state_t >> encoding_int >> alphabet_string;
    TM::Alphabet alphabet( alphabet_string.c_str());
    encoding = static_cast<TM::Encoding>( encoding_int);
    TM_ALWAYS_CHECK(
        ( encoding == TM::ENCODING_F0)
     || ( encoding == TM::ENCODING_F1)
     || ( encoding == TM::ENCODING_F2));
    f.ignore( 256, '\n');

    TM::Turing_machine tm( initial_state_t, 1, alphabet);

    while( true) {
        char c = f.peek();
        if( (c == '#') || (c == '\n')) {
            f.ignore( 256, '\n');
            continue;
        }

        TM::State state;
        TM::Character character;
        TM::State new_state;
        TM::Character new_character;
        TM::Direction direction;

        f >> state >> character >> new_state >> new_character >> direction;
        f.ignore( 256, '\n');
        if( f.eof())
            break;

        tm.add_transition( TM::Transition(
            state, character, new_state, new_character, direction));
    }

    return tm;
}

void run_w(
    const TM::Turing_machine& tm,
    const TM::Character* inputs[],
    const TM::Character* outputs[],
    bool debug = false)
{
    size_t index = 0;
    while( inputs[index]) {

        TM::Turing_machine_instance i( tm, inputs[index]);
        if( debug)
            std::cout << "W input: " << i.get_tape() << std::endl;
        i.run( 0, false);

        if( debug) {
            std::cout << "W output: " << i.get_tape() << std::endl;
            std::cout << "W steps: " << i.get_steps() << std::endl;
        } else {
            std::ostringstream str;
            str << i.get_tape();
            TM_ALWAYS_CHECK( str.str() == outputs[index]);
        }

        ++index;
    }
}

void check_t(
    const TM::Turing_machine& tm_t,
    const TM::Character* inputs[],
    TM::Encoding encoding,
    bool debug = false)
{
    size_t index = 0;
    while( inputs[index]) {

        // Modify TM T for testing:
        // needed because TM T starts with the head right of the input.
        TM::Turing_machine tm_tt = tm_t;
        TM_ALWAYS_CHECK( tm_tt.get_initial_state() != 0);
        tm_tt.set_initial_state( 0);
        tm_tt.add_transition( TM::Transition( 0, '0', 0, '0', 'R'));
        tm_tt.add_transition( TM::Transition( 0, '1', 0, '1', 'R'));
        tm_tt.add_transition(
            TM::Transition( 0, 'B', tm_t.get_initial_state(), 'B', 'N'));

        TM::Turing_machine_instance i( tm_tt, inputs[index]);
        if( debug)
            std::cout << "T input: " << i.get_tape() << std::endl;
        i.run( 0, false);

        if( debug) {
            std::cout << "T output: " << i.get_tape() << std::endl;
            std::cout << "T steps: " << i.get_steps() << std::endl;
        }

        // Check that input for TM T is unaltered.
        std::string input = inputs[index];
        std::ostringstream str;
        str << i.get_tape();
        std::string output = str.str();
        if( encoding == TM::ENCODING_F2) {
            // The prefix length depends on the construction of the transcoder.
            // The following checks assumes that the transcoder shift the input
            // by one cell to the left.
            size_t prefix_length = -i.get_tape().get_left_boundary() - 1;
            TM_ALWAYS_CHECK(
                output.substr( prefix_length, input.size()+2) == input + "11");
        } else {
            TM_ALWAYS_CHECK(
                output.substr( 0, input.size()) == input);
        }

        // Obtain Goedel number of TM W.
        std::string goedel;
        if( encoding == TM::ENCODING_F2)
            // Due to special input the output is the Goedel number of TM W.
            goedel = output;
        else {
            // Replace input of TM W by 1 to obtain Goedel number of TM W.
            goedel = output;
            goedel.replace( 0, input.size(), "1");
        }

        // Construct TM W.
        TM::Alphabet alphabet_w( tm_t.get_alphabet());
        TM::Turing_machine* tm_w = TM::ungoedel( goedel, encoding, alphabet_w);
        TM_ALWAYS_CHECK( tm_w);

        // Just for testing:
        // needed because TM W stops in the initial state of TM T.
        tm_w->add_transition(
            TM::Transition( tm_t.get_initial_state(), 'B', 1, 'B', 'N'));

        const TM::Character* inputs_w[]
            = { "", "0", "1", "00", "01", "10", "11", "000", "001", "010",
                "011", "100", "101", "110", "111", "101010", 0 };
        const TM::Character* outputs_w[]
            = {  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };

        // Test TM W.
        size_t index_w = 0;
        while( inputs_w[index_w]) {
            outputs_w[index_w] = inputs[index];
            ++index_w;
        }
        run_w( *tm_w, inputs_w, outputs_w);

        delete tm_w;
        ++index;
    }
}

void run_self(
    const TM::Turing_machine& tm,
    const TM::Character* inputs[],
    bool debug = false)
{
    size_t index = 0;
    while( inputs[index]) {

        TM::Turing_machine_instance i( tm, inputs[index]);
        if( debug)
            std::cout << "SELF input: " << i.get_tape() << std::endl;
        i.run( 0, false);

        if( debug) {
            std::cout << "SELF steps: " << i.get_steps() << std::endl;
#if 0
            std::ofstream f_self( "self.output");
            f_self << i.get_tape();
            f_self.close();
#endif
        }

        ++index;
    }
}

void test_t(
    const TM::Turing_machine& t,
    TM::State initial_state_w,
    TM::Encoding encoding)
{
    const TM::Character* inputs_01[]
        = { "0", "1", "00", "01", "10", "11", "000", "001", "010", "011",
            "100", "101", "110", "111", "101010", 0 };
    const TM::Character* inputs_2[]
        = { 0, 0 };

    if( encoding == TM::ENCODING_F2) {
        // Transcoders for encoding F2 cannot be tested on arbitrary input
        // The transitions for the states used by the transcoder have to be
        // represented in the input at least as non-existing transitions.
        std::string s;
        for( size_t i = 0; i < initial_state_w-t.get_initial_state()-1; ++i)
            s += "011";
        s += "0"; // The final "11" is constructed from the "##" at the end.
        inputs_2[0] = s.c_str();
        check_t( t, inputs_2, encoding);
    } else
        check_t( t, inputs_01, encoding);
}

void test_self( const TM::Turing_machine& self)
{
    const TM::Character* inputs[] = { "101010", 0 };

    run_self( self, inputs, true);
}

int main( int argc, char* argv[]) {

    TM_ALWAYS_CHECK( argc == 1+1);

    // Read TM T from file and test it.
    TM::State initial_state_w, offset_state_w;
    TM::Encoding encoding;
    TM::Turing_machine tm_t = construct_tm_from_file(
        argv[1], initial_state_w, offset_state_w, encoding);
    print_stats( "T", tm_t, encoding);
    test_t( tm_t, initial_state_w, encoding);

    std::string goedel_t = TM::goedel( tm_t, encoding, true);
    std::cout << "T |goedel|: " << goedel_t.size() << std::endl << std::endl;

    std::string output_w;
    if( encoding == TM::ENCODING_F2) {
        // TM W has to write goedel_t minus the first "111011011" and the
        // last "111".
        output_w = goedel_t.substr( 9);
        output_w.erase( output_w.size()-3);
    } else {
        // TM W has to write goedel_t minus the last "111".
        output_w = goedel_t;
        output_w.erase( output_w.size()-3);
    }

    // Construct TM W.
    TM::Alphabet alphabet_w( tm_t.get_alphabet());
    TM::Turing_machine tm_w( 0, tm_t.get_initial_state(), alphabet_w);
    tm_w.add_transition( TM::Transition( 0, '0', 0, 'B', 'R'));
    tm_w.add_transition( TM::Transition( 0, '1', 0, 'B', 'R'));
    tm_w.add_transition( TM::Transition( 0, 'B', initial_state_w, 'B', 'N'));
    tm_w.add_transitions_for_fixed_string(
        initial_state_w, offset_state_w, output_w.c_str());
    tm_w.add_transition( TM::Transition(
        initial_state_w + offset_state_w*output_w.size(), 'B',
        tm_t.get_initial_state(), 'B', 'N'));
    print_stats( "W", tm_w, encoding);

    std::string goedel_w = TM::goedel( tm_w, encoding, true);
    std::cout << "W |goedel|: " << goedel_w.size() << std::endl << std::endl;

    // Check that TM W and TM T have exactly one state (initial_state_t)
    // in common.
    std::set<TM::State> states_w = tm_w.get_states();
    std::set<TM::State> states_t = tm_t.get_states();
    std::vector<TM::State> states_common( states_w.size() + states_t.size());
    std::vector<TM::State>::iterator it
        = std::set_intersection(
            states_w.begin(), states_w.end(),
            states_t.begin(), states_t.end(),
            states_common.begin());
    TM_ALWAYS_CHECK( it - states_common.begin() == 1);
    TM_ALWAYS_CHECK( states_common[0] == tm_t.get_initial_state());

    // Construct TM SELF.
    TM::Turing_machine tm_self( 0, 1, tm_t.get_alphabet());
    for( size_t index = 0; index < tm_t.get_nr_of_transitions(); ++index)
        tm_self.add_transition( tm_t.get_transition( index));
    for( size_t index = 0; index < tm_w.get_nr_of_transitions(); ++index)
        tm_self.add_transition( tm_w.get_transition( index));
    print_stats( "SELF", tm_self, encoding);

    std::string goedel_self = TM::goedel( tm_self, encoding, true);
    std::cout << "SELF |goedel|: " << size_self << std::endl << std::endl;

#if 0
    std::ofstream f_self( "self.goedel");
    size_t size_self = TM::goedel( f_self, tm_self, encoding);
    f_self.close();
#endif

    test_self( tm_self);

    return 0;
}
