//
// turing - simulation of Turing machines
// Copyright (C) 2009  Joachim Reichel <joachim.reichel@gmx.de>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//

/// \file
/// \brief The #TM::Turing_machine class.

#ifndef TURING_TURING_MACHINE_H
#define TURING_TURING_MACHINE_H

#include "types.h"
#include "alphabet.h"
#include "transition.h"

#include <set>
#include <vector>

namespace TM {



/// Formal description of a Turing machine
///
/// A Turing machine consists of a initial state, a halting state, an
/// alphabet, and a set of transitions.
///
/// Internally, the set of transitions is ordered, i.e., each transition has
/// an index in the range from 0 to #get_nr_of_transitions()-1. The order is
/// important when computing the Goedel number, see #goedel().
class Turing_machine {

public:
    /// Constructor
    Turing_machine( State initial, State halting, const Alphabet& alphabet);

    /// Sets the initial state
    inline void set_initial_state( State initial_state)
    {
        m_initial_state = initial_state;
    };

    /// Returns the initial state
    inline State get_initial_state() const { return m_initial_state; };

    /// Sets the halting state
    inline void set_halting_state( State halting_state)
    {
        m_halting_state = halting_state;
    };

    /// Returns the halting state
    inline State get_halting_state() const { return m_halting_state; };

    /// Returns the alphabet
    inline Alphabet get_alphabet() const { return m_alphabet; }

    /// Adds a transition to the set of transitions
    ///
    /// \pre   The set of transition does not contain a transition with the
    ///        same state and character, i.e., #find_transition() return -1.
    void add_transition( const Transition& transition);

    /// Looks up a particular transition
    ///
    /// Returns the index of the transition where q equals \p state and
    /// c equals \p character.
    ///
    /// \param state       The state of the transition to look up.
    /// \param character   The character of the transition to look up.
    /// \return            The index of the matching transition, or -1 if there
    ///                    is no such transition.
    size_t find_transition(
        const State& state, const Character& character) const;

    /// Returns the number of transitions
    inline size_t get_nr_of_transitions() const
    {
        return m_transition_table.size();
    };

    /// Retuns the transition with a given index
    ///
    /// \pre           \p index < #get_nr_of_transitions()
    ///
    /// \param index   The index of the transition to return.
    /// \return        The transition thas has index \p index.
    inline const Transition& get_transition( size_t index) const
    {
        TM_EXPENSIVE_CHECK( index <= m_transition_table.size());
        return m_transition_table[ index];
    }

    /// Returns the number of states
    size_t get_nr_of_states() const;

    /// Returns the states of this Turing machine
    ///
    /// The states of a Turing machine is the set of all states that appear in
    /// at least one transition. It is not guaranteed that the states can be
    /// reached from the initial state. Note that there is no special handling
    /// of initial and halting state, i.e., they are ignored if they do not
    /// appear in the transitions.
    std::set<State> get_states() const;

    /// Adds transitions that write a fixed string (from left to right)
    ///
    /// The character under the current head position and the next strlen(s)-1
    /// characters to the right must be Blank. The head stops just right of the
    /// last character written.
    ///
    /// \param start_state   The first state to be used by the computed
    ///                      transitions.
    /// \param offset_state  The offset between subsequent states to be used by
    ///                      the computed transitions.
    /// \param s             The fixed strings that is to be written.
    void add_transitions_for_fixed_string(
        size_t start_state, size_t offset_state, const Character* s);

    /// Equality operator
    bool operator==( const Turing_machine& rhs) const;

    /// Inequality operator
    bool operator!=( const Turing_machine& rhs) const;

    /// Output operator
    friend std::ostream& operator<<(
        std::ostream& out, const Turing_machine& turing_machine);

    /// Optimizes internal data structures for improved speed
    ///
    /// This method reorders the transitions by grouping those with the same
    /// state together. An additional map from states to transition indices is
    /// created such that #find_transition() works much faster. The speedup
    /// can be in the order of one magnitude for large Turing machines.
    ///
    /// This method should be called after all transitions have been added.
    /// Adding another transition invalidates the acceleration structure and a
    /// slower, but safe variant is used in the implementation of
    /// #find_transition().
    ///
    /// Note that this method reorders the transitions. That is, their indices
    /// change. The Goedel number changes as well.
    void optimize();

private:
    std::vector<Transition> m_transition_table;
    State m_initial_state;
    State m_halting_state;
    Alphabet m_alphabet;
    std::vector<State> m_state_to_index;
    bool m_state_to_index_map_uptodate;
};

/// Output operator
std::ostream& operator<<(
    std::ostream& out, const Turing_machine& turing_machine);



} // namespace TM

#endif // TURING_TURING_MACHINE_H
