//
// turing - simulation of Turing machines
// Copyright (C) 2009  Joachim Reichel <joachim.reichel@gmx.de>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//

/// \file
/// \brief The #TM::Turing_machine_instance class.

#ifndef TURING_TURING_MACHINE_INSTANCE_H
#define TURING_TURING_MACHINE_INSTANCE_H

#include "types.h"
#include "turing_machine.h"
#include "tape.h"

namespace TM {



/// Concrete instance of a Turing machine
///
/// A concrete instance of a Turing machine can be created from its formal
/// description of type #TM::Turing_machine. In addition to its formal
/// description a concrete instance also has a tape, a current state, and a
/// head position that indiciates the cell of the tape under the head.
///
/// The Turing machine instance also has a counter for the number of performed
/// steps.
class Turing_machine_instance {

public:
    /// Constructor
    ///
    /// The current state is set to the initial state of \p tm. The head
    /// position is set the beginning of the tape content. The step counter
    /// is set to 0.
    ///
    /// \param tm     The formal description of the Turing machine. The
    ///               constructor calls #TM::Turing_machine::optimize() on its
    ///               copy of \p tm.
    /// \param tape   The tape with the input.
    Turing_machine_instance( const Turing_machine& tm, const Tape& tape);

    /// Performs a single step of the Turing machine
    ///
    /// No action is performed if the Turing machine has reached the halting
    /// state.
    ///
    /// \pre   There is a transition that matches the current state and the
    ///        character at the current head position, i.e.,
    ///        #TM::Turing_machine::find_transition() does not return -1, or
    ///        the Turing machine has reached its halting state.
    void step();

    /// Performs several steps of the Turing machine
    ///
    /// The routine repeatedly calls #step() as often as indicated by \p steps.
    /// No more calls are made when the Turing machine reached its halting
    /// state.
    ///
    /// \param steps   The number of steps to perform. A value of 0 indicates
    ///                an unlimited number of steps.
    /// \param trace   If \c true, step the current configuration is printed
    ///                after each setp. Otherwise no output is generated.
    void run( size_t steps = 0, bool trace = false);

    /// Returns the number of performed steps
    inline size_t get_steps() const { return m_steps; }

    /// Returns the current tape
    inline const Tape& get_tape() const { return m_tape; }

    /// Returns the current state
    inline const State& get_state() const { return m_state; }

    /// Checks whether the Turing machine has stopped
    inline bool has_stopped() const
    {
        return m_state == m_turing_machine.get_halting_state();
    }

    /// Equality operator
    bool operator==( const Turing_machine_instance& rhs) const;

    /// Inequality operator
    bool operator!=( const Turing_machine_instance& rhs) const;

    /// Output operator
    friend std::ostream& operator<<(
        std::ostream& out,
        const Turing_machine_instance& turing_machine_instance);

private:
    Turing_machine m_turing_machine;
    Tape           m_tape;
    State          m_state;
    ptrdiff_t      m_head;
    size_t         m_steps;
};

/// Output operator
std::ostream& operator<<(
     std::ostream& out,
     const Turing_machine_instance& turing_machine_instance);



} // namespace TM

#endif // TURING_TURING_MACHINE_INSTANCE_H
