//
// turing - simulation of Turing machines
// Copyright (C) 2009  Joachim Reichel <joachim.reichel@gmx.de>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//

/// \file
/// \brief Basic types and constants.

#ifndef TURING_TYPES_H
#define TURING_TYPES_H

#include <cstring>
#include <cstdlib>

/// The namespace that contains all functionality of this project.
namespace TM {



/// Typdef for states
typedef size_t State;

/// Typedef for characters
typedef char Character;

/// Typedef for directions
///
/// The direction indicates how the head of a TM moves in a transition.
/// There are three possible directions: #Left, #Right, or #Neutral.
typedef char Direction;



/// Blank character
///
/// The blank character is a special character. It may not appear in input
/// and output. All tape cells except those used for the input are labelled
/// with the blank character.
const Character Blank      = 'B';



/// This direction indicates that the head of a TM moves left in a transtion.
const Direction Left       = 'L';

/// This direction indicates that the head of a TM moves right in a transtion.
const Direction Right      = 'R';

/// This direction indicates that the head of a TM does not move in a
/// transtion.
const Direction Neutral    = 'N';



/// Encoding scheme for Goedel numbers
enum Encoding {
    ENCODING_F0 = 0,   //!< The encoding F0, see \ref tm_self_encoding_f0
    ENCODING_F1 = 1,   //!< The encoding F1, see \ref tm_self_encoding_f1
    ENCODING_F2 = 2    //!< The encoding F2, see \ref tm_self_encoding_f2
};



/// Checks whether \p d is a valid direction.
///
/// A direction is valid if it is either #Left, #Right, or #Neutral
inline bool is_valid_direction( Direction d)
{
    return ( d == Left) || ( d == Neutral) || ( d == Right);
}



#define TM_STRINGIFY( X) TM_STRINGIFY1( X)
#define TM_STRINGIFY1( X) TM_STRINGIFY2( X)
#define TM_STRINGIFY2( X) #X

/// Assertion like check used by #TM_ALWAYS_CHECK and #TM_EXPENSIVE_CHECK.
#define TM_CHECK( expr) if( !(expr)) { \
    printf( "%s:%d: %s failed\n", __FILE__, __LINE__, \
    "\"" TM_STRINGIFY( expr) "\""); abort(); }

#ifdef NDEBUG
/// This assertion-like check is always enabled.
#  define TM_ALWAYS_CHECK( expr) TM_CHECK( expr)
/// This assertion-like check is only available in debug builds.
#  define TM_EXPENSIVE_CHECK( expr)
#else
/// This assertion-like check is always enabled.
#  define TM_ALWAYS_CHECK( expr) TM_CHECK( expr)
/// This assertion-like check is only available in debug builds.
#  define TM_EXPENSIVE_CHECK( expr) TM_CHECK( expr)
#endif



} // namespace TM

#endif // TURING_TYPES_H
